from crystallize.experiments.experiment import Experiment
from crystallize.experiments.optimizers import BaseOptimizer, Objective
from crystallize.pipelines.pipeline import Pipeline
from crystallize.experiments.treatment import Treatment
from examples.optimization_experiment import main as opt_example


class MockOptimizer(BaseOptimizer):
    def __init__(self) -> None:
        super().__init__(Objective(metric="sum", direction="minimize"))
        self.ask_count = 0
        self.tell_count = 0

    def ask(self) -> list[Treatment]:
        self.ask_count += 1
        return [Treatment("mock", {"delta": 0})]

    def tell(self, objective_values: dict[str, float]) -> None:
        self.tell_count += 1

    def get_best_treatment(self) -> Treatment:
        return Treatment("best", {"delta": 0})


def test_grid_search_optimizer_runs_and_returns_best() -> None:
    datasource = opt_example.initial_data()
    pipeline = Pipeline([opt_example.add_delta(), opt_example.record_metric()])
    experiment = Experiment(datasource=datasource, pipeline=pipeline)
    optimizer = opt_example.GridSearchOptimizer(
        param_grid={"delta": [0, 1, 2]},
        objective=Objective(metric="sum", direction="minimize"),
    )

    best = experiment.optimize(optimizer, num_trials=3, replicates_per_trial=1)

    assert optimizer.trial_index == 3
    assert optimizer.results == [6, 9, 12]
    ctx = opt_example.FrozenContext({"condition": "test"})
    best.apply(ctx)
    assert ctx.get("delta") == 0


def test_mock_optimizer_counts_calls() -> None:
    datasource = opt_example.initial_data()
    pipeline = Pipeline([opt_example.add_delta(), opt_example.record_metric()])
    experiment = Experiment(datasource=datasource, pipeline=pipeline)
    optimizer = MockOptimizer()

    experiment.optimize(optimizer, num_trials=5, replicates_per_trial=1)

    assert optimizer.ask_count == 5
    assert optimizer.tell_count == 5



import pytest

from crystallize.experiments.result_structs import (
    ExperimentMetrics,
    TreatmentMetrics,
    HypothesisResult,
)


def make_metrics() -> ExperimentMetrics:
    return ExperimentMetrics(
        baseline=TreatmentMetrics({"m": [1]}),
        treatments={"t": TreatmentMetrics({"m": [2]})},
        hypotheses=[
            HypothesisResult(
                name="h",
                results={"t": {"score": 3}},
                ranking={"best": "t"},
            )
        ],
    )


def test_metrics_to_dataframe():
    metrics = make_metrics()
    df = metrics.to_df()
    assert set(df.columns) == {"condition", "hypothesis", "score"}
    assert df.iloc[0]["condition"] == "t"
    assert df.iloc[0]["hypothesis"] == "h"
    assert df.iloc[0]["score"] == 3


def test_metrics_to_df_no_pandas(monkeypatch):
    import crystallize.experiments.result_structs as rs

    metrics = make_metrics()
    monkeypatch.setattr(rs, "pd", None)
    with pytest.raises(ImportError):
        metrics.to_df()



"""Screen for selecting run strategy and artifacts to delete."""

from __future__ import annotations

from pathlib import Path
from typing import List, Tuple

from textual.app import ComposeResult
from textual.containers import Container, Horizontal
from textual.screen import ModalScreen
from textual.widgets import Button, Static, SelectionList
from textual.widgets.selection_list import Selection

from .selection_screens import ActionableSelectionList, SingleSelectionList


class PrepareRunScreen(ModalScreen[tuple[str, tuple[int, ...]] | None]):
    """Collect execution strategy and deletable artifacts."""

    CSS_PATH = "style/prepare_run.tcss"

    BINDINGS = [
        ("ctrl+c", "cancel_and_exit", "Cancel"),
        ("escape", "cancel_and_exit", "Cancel"),
        ("q", "cancel_and_exit", "Close"),
        ("r", "run", "Run"),
    ]

    def __init__(self, deletable: List[Tuple[str, Path]]) -> None:
        super().__init__()
        self._deletable = deletable
        self._strategy: str | None = "rerun"

    def compose(self) -> ComposeResult:
        with Container(id="prepare-run-container"):
            yield Static("Configure Run", id="modal-title")
            yield Static("Press r to run", id="run-hint")

            self.options = SingleSelectionList(
                Selection(
                    "rerun - Will overwrite existing data",
                    "rerun",
                    id="rerun",
                    initial_state=True,
                ),
                Selection("resume - Will skip existing data", "resume", id="resume"),
                id="run-method",
            )
            yield self.options
            if self._deletable:
                yield Static(
                    "Select data to delete (optional)",
                    id="delete-info",
                    classes="invisible",
                )
                self.list = ActionableSelectionList(classes="invisible")
                for idx, (name, path) in enumerate(self._deletable):
                    self.list.add_option(Selection(f"  {name}: {path}", idx))
                yield self.list
            with Horizontal(classes="button-row"):
                yield Button("Run", variant="success", id="run")
                yield Button("Cancel", variant="error", id="cancel")
            yield Static(id="run-feedback")

    def on_mount(self) -> None:
        self.options.focus()

    def on_selection_list_selected_changed(
        self, message: SelectionList.SelectedChanged
    ) -> None:
        if message.selection_list.selected:
            self._strategy = str(message.selection_list.selected[0])
            if self._strategy == "resume" and self._deletable:
                self.list.remove_class("invisible")
                self.query_one("#delete-info").remove_class("invisible")
            elif self._strategy == "rerun" and self._deletable:
                self.query_one("#delete-info").add_class("invisible")
                self.list.add_class("invisible")

    def action_run(self) -> None:
        if self._strategy is None:
            self.query_one("#run-feedback", Static).update(
                f"[red]Select a run strategy to continue[/red]"
            )
            return
        self.dismiss((self._strategy, ()))

    def on_actionable_selection_list_submitted(
        self, message: ActionableSelectionList.Submitted
    ) -> None:
        if self._strategy is not None:
            self.dismiss((self._strategy, message.selected))

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "run":
            if self._strategy is None:
                self.query_one("#run-feedback", Static).update(
                    f"[red]Select a run strategy to continue[/red]"
                )
                return
            selections: tuple[int, ...] = ()
            if hasattr(self, "list"):
                selections = tuple(v for v in self.list.selected if isinstance(v, int))
            self.dismiss((self._strategy, selections))
        else:
            self.dismiss(None)

    def action_cancel_and_exit(self) -> None:
        self.dismiss(None)



from __future__ import annotations

from typing import Dict

from textual.app import ComposeResult
from textual.containers import Container
from textual.widgets import Button, Static, Tree
from textual.screen import ModalScreen


class LoadErrorsScreen(ModalScreen[None]):
    """Display import errors found during discovery."""

    BINDINGS = [
        ("ctrl+c", "close", "Close"),
        ("escape", "close", "Close"),
        ("q", "close", "Close"),
    ]

    def __init__(self, errors: Dict[str, BaseException]) -> None:
        super().__init__()
        self._errors = errors

    def compose(self) -> ComposeResult:
        with Container():
            yield Static("Load Errors", id="modal-title")
            tree = Tree("Failed to load files")
            for file, err in self._errors.items():
                node = tree.root.add(str(file))
                node.add(str(err))
            yield tree
            yield Button("Close", id="close")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.dismiss(None)

    def action_close(self) -> None:
        self.dismiss(None)



from pathlib import Path
import yaml
import pytest

from cli.utils import create_experiment_scaffolding


def test_create_basic(tmp_path: Path) -> None:
    path = create_experiment_scaffolding("exp", directory=tmp_path)
    assert (path / "steps.py").exists()
    assert (path / "datasources.py").exists()
    cfg = yaml.safe_load((path / "config.yaml").read_text())
    assert cfg["name"] == "exp"
    assert cfg["steps"] == []
    assert cfg["datasource"] == {}


def test_create_with_examples(tmp_path: Path) -> None:
    path = create_experiment_scaffolding("demo", directory=tmp_path, examples=True)
    cfg = yaml.safe_load((path / "config.yaml").read_text())
    assert cfg["steps"] == ["add_one"]
    assert cfg["datasource"] == {"numbers": "numbers"}


def test_create_invalid_name(tmp_path: Path) -> None:
    with pytest.raises(ValueError):
        create_experiment_scaffolding("Bad Name", directory=tmp_path)


def test_create_with_artifact_inputs(tmp_path: Path) -> None:
    # upstream experiment providing an output
    up_path = create_experiment_scaffolding("up", directory=tmp_path, outputs=True)
    cfg = yaml.safe_load((up_path / "config.yaml").read_text())
    assert "outputs" in cfg

    # downstream experiment referencing the upstream output
    down = create_experiment_scaffolding(
        "down",
        directory=tmp_path,
        artifact_inputs={"up_out": "up#out"},
    )
    down_cfg = yaml.safe_load((down / "config.yaml").read_text())
    assert down_cfg["datasource"] == {"up_out": "up#out"}




from __future__ import annotations

import importlib
import json
import os
from abc import ABC
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Mapping

from crystallize.utils.constants import (
    REPLICATE_KEY,
    CONDITION_KEY,
    BASELINE_CONDITION,
    METADATA_FILENAME,
)

if TYPE_CHECKING:
    from crystallize.utils.context import FrozenContext
    from crystallize.experiments.experiment import Experiment
    from crystallize.pipelines.pipeline_step import PipelineStep
    from crystallize.experiments.result import Result


def default_seed_function(seed: int) -> None:
    """Set deterministic seeds for common libraries if available."""
    try:
        random_mod = importlib.import_module("random")
        random_mod.seed(seed)
    except ModuleNotFoundError:  # pragma: no cover - stdlib always there in tests
        pass


class BasePlugin(ABC):
    """Interface for extending the :class:`~crystallize.experiments.experiment.Experiment` lifecycle.

    Subclasses can override any of the hook methods to observe or modify the
    behaviour of an experiment.  Hooks are called in a well-defined order during
    :meth:`Experiment.run` allowing plugins to coordinate tasks such as
    seeding, logging, artifact storage or custom execution strategies.
    """

    def init_hook(self, experiment: Experiment) -> None:
        """Configure the experiment instance during initialization."""
        pass

    def before_run(self, experiment: Experiment) -> None:
        """Execute logic before :meth:`Experiment.run` begins."""
        pass

    def before_replicate(self, experiment: Experiment, ctx: FrozenContext) -> None:
        """Run prior to each pipeline execution for a replicate."""
        pass

    def after_step(
        self,
        experiment: Experiment,
        step: PipelineStep,
        data: Any,
        ctx: FrozenContext,
    ) -> None:
        """Observe results after every :class:`PipelineStep` execution."""
        pass

    def after_run(self, experiment: Experiment, result: Result) -> None:
        """Execute cleanup or reporting after :meth:`Experiment.run` completes."""
        pass

    def run_experiment_loop(
        self,
        experiment: "Experiment",
        replicate_fn: Callable[[int], Any],
    ) -> List[Any]:
        """Run all replicates and return their results.

        Returning ``NotImplemented`` signals that the plugin does not provide a
        custom execution strategy and the default should be used instead.
        """
        return NotImplemented


@dataclass
class SeedPlugin(BasePlugin):
    """Manage deterministic seeding for all random operations."""

    seed: Optional[int] = None
    auto_seed: bool = True
    seed_fn: Optional[Callable[[int], None]] = None

    def init_hook(self, experiment: Experiment) -> None:  # pragma: no cover - simple
        pass

    def before_replicate(self, experiment: Experiment, ctx: FrozenContext) -> None:
        if not self.auto_seed:
            return
        local_seed = hash((self.seed or 0) + ctx.get("replicate", 0))
        seed_fn = self.seed_fn or default_seed_function
        seed_fn(local_seed)
        ctx.add("seed_used", local_seed)


@dataclass
class LoggingPlugin(BasePlugin):
    """Configure experiment logging using the ``crystallize`` logger."""

    verbose: bool = False
    log_level: str = "INFO"

    def init_hook(self, experiment: Experiment) -> None:  # pragma: no cover - simple
        pass

    def before_run(self, experiment: Experiment) -> None:
        import logging
        import time

        logger = logging.getLogger("crystallize")
        level = getattr(logging, self.log_level.upper(), logging.INFO)
        logger.setLevel(level)
        if not logger.handlers:
            handler = logging.StreamHandler()
            formatter = logging.Formatter(
                "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
            )
            handler.setFormatter(formatter)
            logger.addHandler(handler)
        seed_plugin = experiment.get_plugin(SeedPlugin)
        seed_val = seed_plugin.seed if seed_plugin else None
        logger.info(
            "Experiment: %d replicates, %d treatments, %d hypotheses (seed=%s)",
            experiment.replicates,
            len(experiment.treatments),
            len(experiment.hypotheses),
            seed_val,
        )
        if seed_plugin and seed_plugin.auto_seed and seed_plugin.seed_fn is None:
            logger.warning("No seed_fn provided—randomness may not be reproducible")
        experiment._start_time = time.perf_counter()

    def after_step(
        self,
        experiment: Experiment,
        step: PipelineStep,
        data: Any,
        ctx: FrozenContext,
    ) -> None:
        if not self.verbose:
            return
        import logging

        logger = logging.getLogger("crystallize")
        logger.info(
            "Rep %s/%s %s finished step %s",
            ctx.get("replicate"),
            experiment.replicates,
            ctx.get("condition"),
            step.__class__.__name__,
        )

    def after_run(self, experiment: Experiment, result: Result) -> None:
        import logging
        import time

        logger = logging.getLogger("crystallize")
        duration = time.perf_counter() - getattr(experiment, "_start_time", 0)
        bests = [
            f"{h.name}: '{h.ranking.get('best')}'"
            for h in result.metrics.hypotheses
            if h.ranking.get("best") is not None
        ]
        best_summary = "; Best " + ", ".join(bests) if bests else ""
        logger.info(
            "Completed in %.1fs%s; %d errors",
            duration,
            best_summary,
            len(result.errors),
        )


@dataclass
class ArtifactPlugin(BasePlugin):
    """Persist artifacts produced during pipeline execution."""

    root_dir: str = "./data"
    versioned: bool = False

    def __post_init__(self) -> None:
        self._manifest: dict[str, str] = {}

    def before_run(self, experiment: Experiment) -> None:
        self.experiment_id = experiment.name or experiment.id
        base = Path(self.root_dir) / self.experiment_id
        base.mkdir(parents=True, exist_ok=True)
        if self.versioned:
            versions = [
                int(p.name[1:])
                for p in base.glob("v*")
                if p.name.startswith("v") and p.name[1:].isdigit()
            ]
            self.version = max(versions, default=-1) + 1
        else:
            self.version = 0
        self._manifest.clear()

    def after_step(
        self,
        experiment: Experiment,
        step: PipelineStep,
        data: Any,
        ctx: FrozenContext,
    ) -> None:
        """Write any artifacts logged in ``ctx.artifacts`` to disk."""
        if len(ctx.artifacts) == 0:
            return
        rep = ctx.get(REPLICATE_KEY, 0)
        condition = ctx.get(CONDITION_KEY, BASELINE_CONDITION)
        for artifact in ctx.artifacts:
            artifact.step_name = step.__class__.__name__
            self._manifest[artifact.name] = artifact.step_name
            dest = (
                Path(self.root_dir)
                / self.experiment_id
                / f"v{self.version}"
                / f"replicate_{rep}"
                / condition
                / artifact.step_name
            )
            os.makedirs(dest, exist_ok=True)
            with open(dest / artifact.name, "wb") as f:
                f.write(artifact.data)
        ctx.artifacts.clear()

    def after_run(self, experiment: Experiment, result: Result) -> None:
        meta = {
            "replicates": experiment.replicates,
            "id": experiment.id,
            "name": experiment.name,
        }
        base = Path(self.root_dir) / self.experiment_id / f"v{self.version}"
        os.makedirs(base, exist_ok=True)
        with open(base / METADATA_FILENAME, "w") as f:
            json.dump(meta, f)

        def dump_condition(name: str, metrics: Mapping[str, Any]) -> None:
            dest = base / name
            os.makedirs(dest, exist_ok=True)
            def _default(o: Any) -> Any:
                try:
                    import numpy as np

                    if isinstance(o, np.ndarray):
                        return o.tolist()
                    if isinstance(o, np.generic):
                        return o.item()
                except Exception:  # pragma: no cover - numpy optional
                    pass
                raise TypeError(
                    f"Object of type {o.__class__.__name__} is not JSON serializable"
                )

            with open(dest / "results.json", "w") as f:
                json.dump({"metrics": metrics}, f, default=_default)
            open(dest / ".crystallize_complete", "a").close()

        dump_condition(BASELINE_CONDITION, result.metrics.baseline.metrics)
        for t_name, m in result.metrics.treatments.items():
            dump_condition(t_name, m.metrics)

        with open(base / "_manifest.json", "w") as f:
            json.dump(self._manifest, f)
        self._manifest.clear()



from crystallize.utils.exceptions import MissingMetricError, PipelineExecutionError


def test_missing_metric_error_message():
    err = MissingMetricError("acc")
    assert "acc" in str(err)


def test_pipeline_execution_error_message():
    original = ValueError("boom")
    err = PipelineExecutionError("MyStep", original)
    assert "MyStep" in str(err)
    assert "boom" in str(err)
    assert err.original_exception is original



"""Screen for displaying run summaries."""

from __future__ import annotations

from typing import Any

from textual.app import ComposeResult
from textual.containers import Container
from textual.widgets import Button, RichLog, Static
from textual.screen import ModalScreen

from ..utils import _write_summary


class SummaryScreen(ModalScreen[None]):
    """Display the summary of an experiment run."""

    BINDINGS = [
        ("ctrl+c", "close", "Close"),
        ("escape", "close", "Close"),
        ("q", "close", "Close"),
    ]

    def __init__(self, result: Any) -> None:
        super().__init__()
        self._result = result

    def compose(self) -> ComposeResult:
        with Container():
            yield Static("Execution Summary", id="modal-title")
            self.log_widget = RichLog()
            yield self.log_widget
            yield Button("Close", id="close")

    async def on_mount(self) -> None:
        _write_summary(self.log_widget, self._result)

    def action_close(self) -> None:
        self.dismiss(None)

    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.dismiss(None)



import asyncio
from pathlib import Path
from typing import Any

import pytest

from cli.discovery import _import_module, _run_object, discover_objects
from cli.utils import (
    _build_experiment_table,
    _build_hypothesis_tables,
    _write_experiment_summary,
    _write_summary,
    filter_mapping,
)
from crystallize import data_source, pipeline_step
from crystallize.experiments.experiment import Experiment
from crystallize.experiments.experiment_graph import ExperimentGraph
from crystallize.experiments.result import Result
from crystallize.experiments.result_structs import (
    ExperimentMetrics,
    TreatmentMetrics,
    HypothesisResult,
)
from crystallize.pipelines.pipeline import Pipeline


class FakeLog:
    def __init__(self) -> None:
        self.written: list[Any] = []

    def write(self, message: Any) -> None:
        self.written.append(message)


@data_source
def dummy_source(ctx):
    return ctx.get("replicate", 0)


@pipeline_step()
def add_one(data, ctx):
    ctx.metrics.add("val", data + 1)
    return {"val": data + 1}


def create_module(tmp: Path) -> Path:
    content = (
        "from crystallize import Experiment, ExperimentGraph, Pipeline, data_source, pipeline_step\n"
        "@data_source\n"
        "def ds(ctx):\n    return 0\n"
        "@pipeline_step()\n"
        "def step(data, ctx):\n    return data\n"
        "\nexp = Experiment(datasource=ds(), pipeline=Pipeline([step()]), name='e')\n"
        "exp.validate()\n"
        "graph = ExperimentGraph.from_experiments([exp])\n"
    )
    mod = tmp / "m.py"
    mod.write_text(content)
    return mod


def make_result() -> Result:
    metrics = ExperimentMetrics(
        baseline=TreatmentMetrics({"m": [0]}),
        treatments={"t": TreatmentMetrics({"m": [1]})},
        hypotheses=[],
    )
    return Result(metrics=metrics, errors={})


def make_hyp_result() -> Result:
    metrics = ExperimentMetrics(
        baseline=TreatmentMetrics({"m": [0]}),
        treatments={"t": TreatmentMetrics({"m": [1]})},
        hypotheses=[
            HypothesisResult(
                name="h",
                results={"t": {"p_value": 0.1, "significant": False}},
                ranking={"best": "t"},
            )
        ],
    )
    return Result(metrics=metrics, errors={})


def test_import_module_relative(tmp_path: Path):
    mod = tmp_path / "mod.py"
    mod.write_text("X = 1")
    m, err = _import_module(mod, tmp_path)
    assert m is not None and getattr(m, "X") == 1 and err is None


def test_import_module_invalid(tmp_path: Path):
    bad = tmp_path / "bad.py"
    bad.write_text("def broken(")
    mod, err = _import_module(bad, tmp_path)
    assert mod is None


def test_import_module_absolute(tmp_path: Path):
    mod = tmp_path / "abs.py"
    mod.write_text("X = 2")
    m, err = _import_module(mod, Path.cwd())
    assert m is not None and getattr(m, "X") == 2 and err is None


def test_import_module_runtime_error(tmp_path: Path):
    bad = tmp_path / "bad.py"
    bad.write_text("raise ValueError('boom')")
    mod, err = _import_module(bad, Path.cwd())
    assert mod is None


def test_discover_objects_skips_invalid(tmp_path: Path):
    create_module(tmp_path)
    bad = tmp_path / "bad.py"
    bad.write_text("def broken(")
    exps, errors = discover_objects(tmp_path, Experiment)
    assert len(exps) == 1


def test_discover_objects_nested(tmp_path: Path):
    pkg = tmp_path / "pkg"
    sub = pkg / "sub"
    sub.mkdir(parents=True)

    mod1 = create_module(pkg)
    mod2 = create_module(sub)

    objs, _ = discover_objects(pkg, ExperimentGraph)
    keys = {Path(k.split(":")[0]) for k in objs}
    assert mod1 in keys and mod2 in keys


def test_run_object_graph_override_replicates():
    exp = Experiment(
        datasource=dummy_source(), pipeline=Pipeline([add_one()]), name="e"
    )
    exp.validate()
    graph = ExperimentGraph.from_experiments([exp])
    results = asyncio.run(_run_object(graph, "rerun", replicates=2))
    assert results["e"].metrics.baseline.metrics["val"] == [1, 2]


def test_run_object_experiment_single():
    exp = Experiment(datasource=dummy_source(), pipeline=Pipeline([add_one()]))
    exp.validate()
    result = asyncio.run(_run_object(exp, "rerun", replicates=2))
    assert result.metrics.baseline.metrics["val"] == [1]


def test_run_object_error_bubbles():
    class BoomExperiment(Experiment):
        async def arun(self, *args: Any, **kwargs: Any) -> Any:  # type: ignore[override]
            raise RuntimeError("boom")

    exp = BoomExperiment(datasource=dummy_source(), pipeline=Pipeline([add_one()]))
    exp.validate()
    with pytest.raises(RuntimeError):
        asyncio.run(_run_object(exp, "rerun", replicates=1))


def test_build_experiment_table():
    table = _build_experiment_table(make_result())
    assert table is not None
    assert list(table.columns[0].cells) == ["m"]
    assert list(table.columns[1].cells) == ["[0]"]
    assert list(table.columns[2].cells) == ["[1]"]


def test_build_experiment_table_multiple_metrics():
    metrics = ExperimentMetrics(
        baseline=TreatmentMetrics({"a": [1], "b": [2]}),
        treatments={
            "t1": TreatmentMetrics({"a": [3], "c": [4]}),
            "t2": TreatmentMetrics({"b": [5]}),
        },
        hypotheses=[],
    )
    table = _build_experiment_table(Result(metrics=metrics, errors={}))
    assert table is not None
    assert [c.header for c in table.columns] == ["Metric", "Baseline", "t1", "t2"]
    rows = [[cell for cell in col.cells] for col in table.columns]
    assert rows[0] == ["a", "b", "c"]
    assert rows[1] == ["[1]", "[2]", "None"]
    assert rows[2] == ["[3]", "None", "[4]"]
    assert rows[3] == ["None", "[5]", "None"]


def test_build_experiment_table_no_metrics():
    metrics = ExperimentMetrics(
        baseline=TreatmentMetrics({}), treatments={}, hypotheses=[]
    )
    result = Result(metrics=metrics, errors={})
    assert _build_experiment_table(result) is None


def test_build_hypothesis_tables():
    res = make_hyp_result()
    tables = _build_hypothesis_tables(res)
    assert len(tables) == 1
    assert tables[0].title.startswith("Hypothesis: h")


def test_write_experiment_summary_with_errors():
    res = make_result()
    res.errors = {"base": RuntimeError("fail")}
    log = FakeLog()
    _write_experiment_summary(log, res)
    assert any(isinstance(m, type(_build_experiment_table(res))) for m in log.written)
    assert any("Errors occurred" in str(m) for m in log.written)
    assert any("base:" in str(m) for m in log.written)


def test_write_experiment_summary_errors_only():
    metrics = ExperimentMetrics(
        baseline=TreatmentMetrics({}), treatments={}, hypotheses=[]
    )
    res = Result(metrics=metrics, errors={"boom": RuntimeError("fail")})
    log = FakeLog()
    _write_experiment_summary(log, res)
    assert not any(hasattr(m, "columns") for m in log.written)
    assert any("Errors occurred" in str(m) for m in log.written)


def test_write_experiment_summary_with_hypotheses():
    res = make_hyp_result()
    log = FakeLog()
    _write_experiment_summary(log, res)
    titles = [getattr(m, "title", "") for m in log.written if hasattr(m, "title")]
    assert "Metrics" in titles
    assert any(isinstance(t, str) and t.startswith("Hypothesis: h") for t in titles)


def test_write_summary_single_result():
    res = make_result()
    log = FakeLog()
    _write_summary(log, res)
    assert any(hasattr(m, "columns") for m in log.written)
    assert not any(str(m) == "exp" for m in log.written)


def test_write_summary_nested_dicts():
    res1 = make_result()
    res2 = make_result()
    log = FakeLog()
    _write_summary(log, {"e1": res1, "e2": res2})
    heads = [str(m) for m in log.written if "Errors" not in str(m)]
    assert any("e1" in h for h in heads)
    assert any("e2" in h for h in heads)


def test_write_summary_dict():
    res = make_result()
    log = FakeLog()
    _write_summary(log, {"exp": res})
    assert any(str(m) == "exp" for m in log.written if not hasattr(m, "columns"))
    assert any(hasattr(m, "columns") for m in log.written)


def test_filter_mapping():
    data = {"ExpOne": 1, "Another": 2, "expTwo": 3}
    assert filter_mapping(data, "exp") == {"ExpOne": 1, "expTwo": 3}



"""Reusable selection widgets for the CLI."""

from __future__ import annotations

from typing import Any

from textual.message import Message
from textual.widgets import SelectionList
from textual.widgets.selection_list import Selection


class ActionableSelectionList(SelectionList):
    """A SelectionList that emits a Submitted message on Enter."""

    class Submitted(Message):
        def __init__(self, selected: tuple[Any, ...]) -> None:
            self.selected = selected
            super().__init__()

    BINDINGS = [("escape", "cancel", "Cancel")]

    def action_submit(self) -> None:
        selected_indices = tuple(
            value for value in self.selected if isinstance(value, int)
        )
        self.post_message(self.Submitted(selected_indices))


class SingleSelectionList(ActionableSelectionList):
    """A SelectionList that allows only a single selection."""

    def select(self, selection: Selection | Any) -> "SingleSelectionList":
        self.deselect_all()
        return super().select(selection)

    def toggle(self, selection: Selection | Any) -> "SingleSelectionList":
        if selection in self.selected or (
            isinstance(selection, Selection) and selection.value in self.selected
        ):
            self.deselect_all()
        else:
            self.deselect_all()
            super().select(selection)
        self.refresh()
        return self



"""Object discovery utilities for the CLI."""

from __future__ import annotations

import importlib.util
import inspect
import sys
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Type

import yaml

from crystallize.experiments.experiment_graph import ExperimentGraph


def _import_module(
    file_path: Path, root_path: Path
) -> Tuple[Optional[Any], Optional[BaseException]]:
    """Import ``file_path`` as a module relative to ``root_path``.

    Returns a tuple of the imported module (or ``None`` if import failed) and
    the exception raised during import, if any.
    """
    try:
        relative_path = file_path.relative_to(root_path)
    except ValueError:
        spec = importlib.util.spec_from_file_location(file_path.stem, file_path)
        if spec and spec.loader:
            module = importlib.util.module_from_spec(spec)
            try:
                spec.loader.exec_module(module)  # type: ignore[arg-type]
                return module, None
            except BaseException as exc:  # noqa: BLE001
                return None, exc
        return None, RuntimeError("Could not create module spec")

    module_name = ".".join(relative_path.with_suffix("").parts)

    try:
        if str(root_path) not in sys.path:
            sys.path.insert(0, str(root_path))

        if module_name in sys.modules:
            # If module is already imported, reload it to pick up changes
            return importlib.reload(sys.modules[module_name]), None
        else:
            # Otherwise, import it for the first time
            return importlib.import_module(module_name), None
    except BaseException as exc:  # noqa: BLE001
        return None, exc


def discover_objects(
    directory: Path, obj_type: Type[Any]
) -> Tuple[Dict[str, Any], Dict[str, BaseException]]:
    """Recursively discover objects of ``obj_type`` within ``directory``.

    Returns a mapping of discovered objects and a mapping of file paths to
    import errors for modules that failed to load.
    """
    abs_directory = directory.resolve()
    root_path = Path.cwd()
    found: Dict[str, Any] = {}
    errors: Dict[str, BaseException] = {}
    for file in abs_directory.rglob("*.py"):
        mod, err = _import_module(file, root_path)
        if err is not None:
            errors[str(file)] = err
        if not mod:
            continue
        for name, obj in inspect.getmembers(mod, lambda x: isinstance(x, obj_type)):
            try:
                rel = file.relative_to(root_path)
            except ValueError:
                rel = file
            found[f"{rel}:{name}"] = obj
    return found, errors


def discover_configs(
    directory: Path,
) -> Tuple[
    Dict[str, Dict[str, Any]],
    Dict[str, Dict[str, Any]],
    Dict[str, BaseException],
]:
    """Discover experiments and graphs defined via ``config.yaml``."""

    def _has_ref(val: Any) -> bool:
        if isinstance(val, str):
            return "#" in val
        if isinstance(val, dict):
            return any(_has_ref(v) for v in val.values())
        if isinstance(val, list):
            return any(_has_ref(v) for v in val)
        return False

    graphs: Dict[str, Dict[str, Any]] = {}
    experiments: Dict[str, Dict[str, Any]] = {}
    errors: Dict[str, BaseException] = {}

    abs_directory = directory.resolve()
    cwd = Path.cwd()

    for cfg in abs_directory.rglob("config.yaml"):
        try:
            with cfg.open() as f:
                data = yaml.safe_load(f) or {}
            name = data.get("name", cfg.parent.name)
            desc = data.get("description", "")
            repl = data.get("replicates", 1)
            cli_cfg = data.get("cli", {}) or {}

            is_graph = _has_ref(data.get("datasource"))

            if cli_cfg.get("hidden"):
                continue

            try:
                rel = cfg.parent.relative_to(cwd)
            except ValueError:
                rel = cfg.parent
            label = f"{rel} - {name}"

            cli_defaults = {
                "group": "Graphs" if is_graph else "Experiments",
                "priority": 999,
                "icon": "📈" if is_graph else "🧪",
                "color": None,
                "hidden": False,
            }
            cli_info = {**cli_defaults, **cli_cfg}

            info = {
                "path": cfg,
                "description": desc,
                "replicates": repl,
                "label": label,
                "cli": cli_info,
            }

            if is_graph:
                graphs[label] = info
            else:
                experiments[label] = info
        except BaseException as exc:  # noqa: BLE001
            errors[str(cfg)] = exc

    return graphs, experiments, errors


async def _run_object(obj: Any, strategy: str, replicates: Optional[int]) -> Any:
    """Run an ``Experiment`` or ``ExperimentGraph`` asynchronously."""
    if isinstance(obj, ExperimentGraph):
        return await obj.arun(strategy=strategy, replicates=replicates)
    return await obj.arun(
        strategy=strategy,
        replicates=None,
        treatments=getattr(obj, "treatments", None),
        hypotheses=getattr(obj, "hypotheses", None),
    )



from __future__ import annotations

from pathlib import Path
import yaml

from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical, VerticalScroll
from textual.reactive import reactive
from textual.screen import ModalScreen
from textual.widgets import (
    Button,
    Checkbox,
    Collapsible,
    Input,
    Label,
    Static,
    Tree,
)
from textual.binding import Binding
from textual.widgets.selection_list import Selection

from ..utils import create_experiment_scaffolding
from .selection_screens import ActionableSelectionList


class OutputTree(Tree):
    """Tree widget with custom binding for output selection."""

    BINDINGS = [b for b in Tree.BINDINGS if getattr(b, "key", "") != "enter"] + [
        Binding("enter", "toggle_output", "Select", show=True)
    ]

    def action_toggle_output(self) -> None:  # pragma: no cover - delegates
        screen = self.screen
        if screen is not None and hasattr(screen, "action_toggle_output"):
            screen.action_toggle_output()

        try:
            line = self._tree_lines[self.cursor_line]
        except IndexError:
            pass
        else:
            node = line.path[-1]
            self.post_message(Tree.NodeSelected(node))


class CreateExperimentScreen(ModalScreen[None]):
    """Interactive screen for creating a new experiment folder."""

    CSS_PATH = "style/create_experiment.tcss"

    BINDINGS = [
        ("ctrl+c", "cancel", "Cancel"),
        ("escape", "cancel", "Cancel"),
        ("q", "cancel", "Close"),
        ("c", "create", "Create"),
        ("enter", "toggle_output", "Select"),
    ]

    name_valid = reactive(False)

    def _fmt_label(self, exp: str, out: str) -> str:
        marker = "[X]" if out in self._selected.get(exp, set()) else "[ ]"
        return f"{marker} {out}"

    def compose(self) -> ComposeResult:
        with VerticalScroll(id="create-exp-container"):
            yield Static("Create New Experiment", id="modal-title")
            yield Input(
                placeholder="Enter experiment name (lowercase, no spaces)",
                id="name-input",
            )
            yield Label(
                "[dim]Enter a name to continue[/dim]", id="name-feedback"
            )  # For validation feedback
            with Horizontal(classes="button-row"):
                yield Checkbox(
                    "Use outputs from other experiments",
                    id="graph-mode",
                )
                yield Checkbox(
                    "Add example code",
                    id="examples",
                    tooltip="Includes starter code in selected files (steps.py, hypotheses.py, etc.)",
                )

            with Collapsible(title="Files to include", collapsed=False):
                self.file_list = ActionableSelectionList(classes="files-to-include")
                self.file_list.add_option(
                    Selection(
                        "steps.py",
                        "steps",
                        initial_state=True,
                        id="steps",
                        disabled=False,
                    )
                )
                self.file_list.add_option(
                    Selection(
                        "datasources.py",
                        "datasources",
                        initial_state=True,
                        id="datasources",
                    )
                )
                self.file_list.add_option(
                    Selection(
                        "outputs.py",
                        "outputs",
                        id="outputs",
                    )
                )
                self.file_list.add_option(
                    Selection(
                        "hypotheses.py",
                        "hypotheses",
                        id="hypotheses",
                    )
                )
                yield self.file_list

            with Vertical(id="graph-container", classes="invisible"):
                with Collapsible(title="Select outputs to use", collapsed=False):
                    self.out_tree = OutputTree("root", id="out-tree")
                    self.out_tree.show_root = False
                    yield self.out_tree

            with Horizontal(classes="button-row"):
                yield Button("Create", variant="success", id="create")
                yield Button("Cancel", variant="error", id="cancel")

    def on_mount(self) -> None:
        self.query_one("#name-input", Input).focus()
        base = Path("experiments")
        self._outputs: dict[str, list[str]] = {}
        self._selected: dict[str, set[str]] = {}
        if base.exists():
            for p in base.iterdir():
                cfg = p / "config.yaml"
                if cfg.exists():
                    with open(cfg) as f:
                        data = yaml.safe_load(f) or {}
                    outs = list((data.get("outputs") or {}).keys())
                    if outs:
                        self._outputs[p.name] = outs
        if hasattr(self, "out_tree"):
            for name in sorted(self._outputs):
                node = self.out_tree.root.add(name)
                for out in self._outputs[name]:
                    leaf = node.add_leaf(self._fmt_label(name, out))
                    leaf.data = (name, out)

    def on_input_changed(self, event: Input.Changed) -> None:
        name = event.value.strip()
        feedback = self.query_one("#name-feedback", Label)
        if not name:
            feedback.update("[dim]Enter a name to continue[/dim]")
            self.name_valid = False
        elif not name.islower() or " " in name:
            feedback.update("[red]Name must be lowercase with no spaces[/red]")
            self.name_valid = False
        else:
            feedback.update(f"[green]Path: experiments/{name}[/green]")
            self.name_valid = True

    def on_checkbox_changed(self, event: Checkbox.Changed) -> None:
        if event.checkbox.id == "graph-mode":
            container = self.query_one("#graph-container")
            if event.value:
                container.remove_class("invisible")
            else:
                container.add_class("invisible")

    def action_toggle_output(self) -> None:
        if hasattr(self, "out_tree"):
            node = self.out_tree.cursor_node
            if node is not None and node.data is not None:
                exp, out = node.data
                selected = self._selected.setdefault(exp, set())
                if out in selected:
                    selected.remove(out)
                else:
                    selected.add(out)
                node.set_label(self._fmt_label(exp, out))

    def action_cancel(self) -> None:
        self.dismiss(None)

    def action_create(self) -> None:
        if self.name_valid:
            self._create()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "create" and self.name_valid:
            self._create()
        else:
            return

    def _create(self) -> None:
        name = self.query_one("#name-input", Input).value.strip()
        base = Path("experiments")
        selections = set(self.file_list.selected)
        examples = self.query_one("#examples", Checkbox).value
        artifact_inputs = {}
        if self.query_one("#graph-mode", Checkbox).value:
            for exp, outs in self._selected.items():
                for out in outs:
                    alias = f"{exp}_{out}" if out in artifact_inputs else out
                    artifact_inputs[alias] = f"{exp}#{out}"
        try:
            create_experiment_scaffolding(
                name,
                directory=base,
                steps="steps" in selections,
                datasources="datasources" in selections,
                outputs="outputs" in selections,
                hypotheses="hypotheses" in selections,
                examples=examples,
                artifact_inputs=artifact_inputs or None,
            )
        except FileExistsError:
            self.app.bell()
            return
        self.dismiss(None)



"""Screens for confirming deletion of artifacts."""

from __future__ import annotations

from pathlib import Path


from textual.app import ComposeResult
from textual.containers import Container, Horizontal, VerticalScroll
from textual.widgets import Button, Static
from textual.screen import ModalScreen



class ConfirmScreen(ModalScreen[bool]):
    BINDINGS = [
        ("ctrl+c", "cancel_and_exit", "Cancel"),
        ("y", "confirm_and_exit", "Confirm"),
        ("n", "cancel_and_exit", "Cancel"),
    ]

    def __init__(self, paths_to_delete: list[Path]) -> None:
        super().__init__()
        self._paths = paths_to_delete

    def compose(self) -> ComposeResult:
        with Container(classes="confirm-delete-container"):
            yield Static(
                "[bold red]The following will be permanently deleted:[/bold red]"
            )
            with VerticalScroll(classes="path-list"):
                if not self._paths:
                    yield Static("  (Nothing selected)")
                for path in self._paths:
                    yield Static(f"• {path}")
            yield Static("\nAre you sure you want to proceed? (y/n)")
            yield Horizontal(
                Button("Yes, Delete", variant="error", id="yes"),
                Button("No, Cancel", variant="primary", id="no"),
            )

    def on_mount(self) -> None:
        self.query_one("#no", Button).focus()

    def action_confirm_and_exit(self) -> None:
        self.dismiss(True)

    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.dismiss(event.button.id == "yes")

    def action_cancel_and_exit(self) -> None:
        self.dismiss(False)



"""Public convenience API for Crystallize."""

from __future__ import annotations

from crystallize.experiments.experiment_graph import ExperimentGraph
from crystallize.experiments.experiment import Experiment
from crystallize.experiments.experiment_builder import ExperimentBuilder
from crystallize.datasources.datasource import DataSource, ExperimentInput
from crystallize.datasources import Artifact
from crystallize.experiments.treatment import Treatment
from crystallize.utils.context import FrozenContext
from crystallize.pipelines.pipeline import PipelineStep
from crystallize.pipelines.pipeline import Pipeline
from crystallize.experiments.hypothesis import Hypothesis

from crystallize.utils.decorators import (
    data_source,
    hypothesis,
    inject_from_ctx,
    pipeline,
    pipeline_step,
    resource_factory,
    treatment,
    verifier,
)
from crystallize.plugins.execution import (
    AsyncExecution,
    ParallelExecution,
    SerialExecution,
)
from crystallize.plugins.plugins import (
    ArtifactPlugin,
    BasePlugin,
    LoggingPlugin,
    SeedPlugin,
)

__all__ = [
    "pipeline_step",
    "inject_from_ctx",
    "treatment",
    "hypothesis",
    "data_source",
    "verifier",
    "pipeline",
    "resource_factory",
    "PipelineStep",
    "Pipeline",
    "Hypothesis",
    "Experiment",
    "DataSource",
    "FrozenContext",
    "Treatment",
    "BasePlugin",
    "SerialExecution",
    "ParallelExecution",
    "AsyncExecution",
    "SeedPlugin",
    "LoggingPlugin",
    "ArtifactPlugin",
    "ExperimentGraph",
    "Artifact",
    "ExperimentInput",
    "ExperimentBuilder",
]



from typing import Any, Callable, Dict, Mapping, Sequence, Optional

from crystallize.utils.exceptions import MissingMetricError


def rank_by_p_value(result: dict) -> float:
    """A simple, picklable ranker function. Lower p-value is better."""
    return result.get("p_value", 1.0)


class Hypothesis:
    """Encapsulate a statistical test to compare baseline and treatment results."""

    def __init__(
        self,
        verifier: Callable[
            [Mapping[str, Sequence[Any]], Mapping[str, Sequence[Any]]],
            Mapping[str, Any],
        ],
        metrics: str | Sequence[str] | Sequence[Sequence[str]] | None = None,
        ranker: Callable[[Mapping[str, Any]], float] | None = None,
        name: Optional[str] = None,
    ) -> None:
        self.metrics_spec = metrics
        self.verifier = verifier
        self.ranker = ranker or rank_by_p_value
        if name:
            self.name = name
        elif hasattr(ranker, "__name__"):
            self.name = ranker.__name__  # type: ignore[arg-type]
        else:
            self.name = "hypothesis"

    # ---- public API -----------------------------------------------------

    def verify(
        self,
        baseline_metrics: Mapping[str, Sequence[Any]],
        treatment_metrics: Mapping[str, Sequence[Any]],
    ) -> Any:
        """Evaluate the hypothesis using selected metric groups.

        Args:
            baseline_metrics: Aggregated metrics from baseline runs.
            treatment_metrics: Aggregated metrics from a treatment.

        Returns:
            The output of the ``verifier`` callable. When multiple metric groups
            are specified the result is a list of outputs in the same order.
        """

        def subset(
            keys: Sequence[str],
        ) -> tuple[Dict[str, Sequence[Any]], Dict[str, Sequence[Any]]]:
            b: Dict[str, Sequence[Any]] = {}
            t: Dict[str, Sequence[Any]] = {}
            for k in keys:
                try:
                    b[k] = baseline_metrics[k]
                    t[k] = treatment_metrics[k]
                except KeyError:
                    raise MissingMetricError(k)
            return b, t

        spec = self.metrics_spec
        if spec is None:
            groups = [list(baseline_metrics.keys())]
        elif isinstance(spec, str):
            groups = [[spec]]
        elif (
            spec
            and isinstance(spec[0], Sequence)
            and not isinstance(spec[0], (str, bytes))
        ):  # type: ignore[index]
            groups = [list(g) for g in spec]  # type: ignore[arg-type]
        else:
            groups = [list(spec)]  # type: ignore[arg-type]

        outputs = []
        for group in groups:
            b, t = subset(group)
            outputs.append(self.verifier(b, t))

        return outputs[0] if len(outputs) == 1 else outputs

    def rank_treatments(self, verifier_results: Mapping[str, Any]) -> Mapping[str, Any]:
        """Rank treatments using the ``ranker`` score function."""
        scores = {name: self.ranker(res) for name, res in verifier_results.items()}
        ranked = sorted(scores.items(), key=lambda x: x[1])
        best = ranked[0][0] if ranked else None
        return {"best": best, "ranked": ranked}



import types
import pytest
from textual.app import App
from textual.widgets import RichLog, Button, TextArea
from cli.screens.run import RunScreen
from crystallize import data_source, pipeline_step
from crystallize.experiments.experiment import Experiment
from crystallize.pipelines.pipeline import Pipeline


@data_source
def ds(ctx):
    return 0


@pipeline_step()
def step(data, ctx):
    return data


@pytest.mark.asyncio
async def test_run_screen_toggle_plain_text():
    exp = Experiment(datasource=ds(), pipeline=Pipeline([step()]))
    exp.validate()
    screen = RunScreen(exp, "rerun", None)

    screen.run_worker = lambda *a, **k: types.SimpleNamespace(
        is_finished=True, cancel=lambda: None
    )

    async with App().run_test() as pilot:
        await pilot.app.push_screen(screen)
        log = screen.query_one("#live_log", RichLog)
        text_area = screen.query_one("#plain_log", TextArea)
        button = screen.query_one("#toggle_text", Button)

        # Initially RichLog visible, TextArea hidden
        assert log.display
        assert not text_area.display
        assert button.label == "Plain Text"

        # Toggle to plain text mode
        await pilot.press("t")
        assert not log.display
        assert text_area.display
        assert button.label == "Rich Text"

        # Toggle back to rich text mode
        await pilot.press("t")
        assert log.display
        assert not text_area.display
        assert button.label == "Plain Text"



from crystallize.utils.cache import compute_hash
from typing import Any
from concurrent.futures import ThreadPoolExecutor
from crystallize.utils.context import FrozenContext
from crystallize.pipelines.pipeline import Pipeline
from crystallize.pipelines.pipeline_step import PipelineStep
import pytest
import numpy as np


class CountingStep(PipelineStep):
    cacheable = True

    def __init__(self):
        self.calls = 0

    def __call__(self, data, ctx):
        self.calls += 1
        return data + 1

    @property
    def params(self):
        return {}


class MetricsStep(PipelineStep):
    cacheable = True

    def __call__(self, data, ctx):
        return {"result": data}

    @property
    def params(self):
        return {}


def test_cache_hit_and_miss(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)

    step = CountingStep()
    pipeline = Pipeline([step, MetricsStep()])
    ctx = FrozenContext({})

    result1 = pipeline.run(0, ctx)
    assert result1 == {"result": 1}
    assert step.calls == 1

    step2 = CountingStep()
    pipeline2 = Pipeline([step2, MetricsStep()])
    result2 = pipeline2.run(0, ctx)
    assert result2 == {"result": 1}
    assert step2.calls == 0
    assert pipeline2.get_provenance()[0]["cache_hit"] is True

    result3 = pipeline2.run(5, ctx)
    assert result3 == {"result": 6}
    assert step2.calls == 1


class NoCacheStep(CountingStep):
    cacheable = False


def test_non_cacheable_step(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)

    step = NoCacheStep()
    pipeline = Pipeline([step, MetricsStep()])
    ctx = FrozenContext({})

    pipeline.run(0, ctx)
    pipeline.run(0, ctx)

    assert step.calls == 2


def test_corrupted_cache_recovers(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)

    step = CountingStep()
    pipeline = Pipeline([step, MetricsStep()])
    ctx = FrozenContext({})

    pipeline.run(0, ctx)

    step_hash = step.step_hash
    input_hash = compute_hash(0)
    cache_file = tmp_path / ".cache" / step_hash / f"{input_hash}.pkl"
    cache_file.write_text("corrupted")

    step.calls = 0
    pipeline.run(0, ctx)

    assert step.calls == 1


class ReturnStep(PipelineStep):
    cacheable = True

    def __init__(self, value: Any):
        self.value = value

    def __call__(self, data, ctx):
        return self.value

    @property
    def params(self):
        return {}


class Unpickleable:
    def __getstate__(self):  # pragma: no cover - can't pickle
        raise TypeError("unpickleable")


@pytest.mark.parametrize(
    "value,should_fail",
    [(np.zeros((10, 10)), False), (Unpickleable(), True)],
)
def test_cache_large_and_unpickleable(tmp_path, monkeypatch, value, should_fail):
    monkeypatch.chdir(tmp_path)
    step1 = ReturnStep(value)
    pipeline1 = Pipeline([step1, MetricsStep()])
    ctx = FrozenContext({})

    if should_fail:
        with pytest.raises(IOError):
            pipeline1.run(None, ctx)
    else:
        pipeline1.run(None, ctx)
        step2 = ReturnStep(value)
        pipeline2 = Pipeline([step2, MetricsStep()])
        pipeline2.run(None, ctx)
        assert pipeline2.get_provenance()[0]["cache_hit"] is True


def test_cache_dir_permission_error(tmp_path, monkeypatch):
    cache_dir = tmp_path / "cache"
    monkeypatch.setenv("CRYSTALLIZE_CACHE_DIR", str(cache_dir))
    monkeypatch.chdir(tmp_path)

    def deny_open(*args, **kwargs):
        raise PermissionError("read-only")

    monkeypatch.setattr("pathlib.Path.open", deny_open)

    step = CountingStep()
    pipeline = Pipeline([step, MetricsStep()])
    ctx = FrozenContext({})

    with pytest.raises(IOError):
        pipeline.run(0, ctx)


class ParamStep(PipelineStep):
    cacheable = True

    def __init__(self, value: int) -> None:
        self.value = value

    def __call__(self, data, ctx):
        return data + self.value

    @property
    def params(self):
        return {"value": self.value}


def test_step_and_input_hash_uniqueness():
    step1 = ParamStep(1)
    step2 = ParamStep(2)
    assert step1.step_hash != step2.step_hash
    assert compute_hash({"x": 1}) != compute_hash({"x": 2})


def test_concurrent_cache_writes(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)

    def run(pipe: Pipeline) -> None:
        ctx = FrozenContext({})
        pipe.run(0, ctx)

    pipes = [Pipeline([CountingStep(), MetricsStep()]) for _ in range(5)]
    with ThreadPoolExecutor(max_workers=5) as ex:
        list(ex.map(run, pipes))

    step = CountingStep()
    pipeline = Pipeline([step, MetricsStep()])
    ctx = FrozenContext({})
    pipeline.run(0, ctx)
    assert step.calls == 0



"""Constants used across the CLI."""

from __future__ import annotations

from crystallize.experiments.experiment import Experiment
from crystallize.experiments.experiment_graph import ExperimentGraph

OBJ_TYPES = {
    "experiment": Experiment,
    "graph": ExperimentGraph,
}

ASCII_ART = r"""
   *           *      _        _ _ *
   ___ _*__ _   _ ___| |_ __ *| | (_)_______ *
  * __| ' _| | | / *_| __/ _` | | | |_* / _ \
 | (__| |* | |_| \__ \ ||*(_| | * | |/ /  __/ *
  \___|_|   *__, |___/\__\__,_|_|_|_/___\___|
    *       |___/ *              *
"""

ASCII_ART_2 = r"""
 ▗▄▄▖▗▄▄▖▗▖  ▗▖▗▄▄▖▗▄▄▄▖▗▄▖ ▗▖   ▗▖   ▗▄▄▄▖▗▄▄▄▄▖▗▄▄▄▖
▐▌   ▐▌ ▐▌▝▚▞▘▐▌     █ ▐▌ ▐▌▐▌   ▐▌     █     ▗▞▘▐▌
▐▌   ▐▛▀▚▖ ▐▌  ▝▀▚▖  █ ▐▛▀▜▌▐▌   ▐▌     █   ▗▞▘  ▐▛▀▀▘
▝▚▄▄▖▐▌ ▐▌ ▐▌ ▗▄▄▞▘  █ ▐▌ ▐▌▐▙▄▄▖▐▙▄▄▖▗▄█▄▖▐▙▄▄▄▖▐▙▄▄▖
"""

ASCII_ART_3 = r"""
┌─┐┬─┐┬ ┬┌─┐┌┬┐┌─┐┬  ┬  ┬┌─┐┌─┐
│  ├┬┘└┬┘└─┐ │ ├─┤│  │  │┌─┘├┤
└─┘┴└─ ┴ └─┘ ┴ ┴ ┴┴─┘┴─┘┴└─┘└─┘
"""

ASCII_ART_4 = r"""
  ___  ____  _  _  ____  ____  __   __    __    __  ____  ____
 / __)(  _ \( \/ )/ ___)(_  _)/ _\ (  )  (  )  (  )(__  )(  __)
( (__  )   / )  / \___ \  )( /    \/ (_/\/ (_/\ )(  / _/  ) _)
 \___)(__\_)(__/  (____/ (__)\_/\_/\____/\____/(__)(____)(____)
"""

# ASCII_ART_ARRAY = [ASCII_ART, ASCII_ART_2, ASCII_ART_3, ASCII_ART_4]
ASCII_ART_ARRAY = [ASCII_ART]

CSS = """
App {
    background: $background;
}

Header {
    content-align: center middle;
}

Static#title {
    content-align: center middle;
    text-style: bold;
    color: $primary;
}

ListView {
    background: $panel;
    border: tall $secondary;
    height: 1fr;
}

ListItem {
    padding: 1;
}

ListItem > Static {
    color: $text;
}

ListItem:hover {
    background: $accent;
}

ListItem.selected {
    background: $success;
}

.experiment-item > Static {
}

.graph-item > Static {
    text-style: bold italic;
}

ModalScreen {
    background: $background 50%;
    align: center middle;
}

PrepareRunScreen Container {
    border: round $warning;
    background: $panel;
    width: 80%;
    height: auto;
    padding: 1;
}

.confirm-delete-container {
    width: 80%;
    height: auto;
    max-height: 80%;
    border: round $error;
    background: $panel;
    padding: 1;
}

.create-exp-container {
    width: 60%;
    height: auto;
    border: round $primary;
    background: $panel;
    padding: 1;
}

.path-list {
    background: $surface;
    border: round $primary;
    margin: 1 0;
    padding: 1;
    height: auto;
    max-height: 10; /* Makes the list scrollable if it's long */
}

SelectionList {
    height: 1fr;
    width: 100%;
}

#dag-display {
    height: auto;
    border: round $primary;
    padding: 0 1;
    margin-bottom: 1;
}

RichLog {
    height: 1fr;
    width: 100%;
    background: $panel;
    border: tall $secondary;
}

Button {
    margin: 1 0;
}

Button#confirm {
    background: $error;
}

Button#yes {
    background: $success;
}

Button#no {
    background: $error;
}

.button-row {
    align-horizontal: center;
}

LoadingIndicator {
    color: $accent;
}

#main-container {
    layout: vertical;
    padding: 1;
}

Horizontal {
    height: 1fr;
}

.left-panel {
    width: 50%;
    height: 1fr;
    border-right: solid $secondary;
    padding-right: 1;
}

.right-panel {
    width: 50%;
    padding-left: 1;
    height: 1fr;

}

.details-panel {
    background: $panel-darken-1;
    border: round $secondary;
    padding: 1;
    height: 1fr;
    content-align: left top;
    overflow: auto;
}

TabbedContent {
    height: 1fr;
}

TabPane {
    height: 1fr;
    padding: 1;
}

ListView {
    border: solid $secondary;
    height: 1fr;
}

ListItem {
    padding: 1;
    layout: vertical;
}

.item-doc {
    color: $text-muted;
}

ListItem:hover {
    background: $accent-lighten-1;
}

.experiment-item:hover, .graph-item:hover {
    background: $accent-lighten-1;
}

#error-msg {
    color: $error;
    margin-top: 1;
}

Input {
    border: tall $secondary;
    margin-bottom: 1;
    width: 100%;
}

#replicate-input {
    width: 10;
    margin-right: 1;
    margin-top: 0;
    align-vertical: middle;
}

Input:focus {
    border: tall $accent;
}

.invisible {
    display: none;
}

#button-container {
    height: auto;
    layout: horizontal;
    margin: 1 0;
    align-vertical: middle;
}

#select-button-container {
    height: auto;
    layout: horizontal;
    align-vertical: middle;
}

#run-btn {
    align-vertical: middle;
    margin-top: 0;
}
"""



from .experiment import Experiment
from .experiment_graph import ExperimentGraph
from .hypothesis import Hypothesis
from .optimizers import BaseOptimizer, Objective
from .result import Result
from .result_structs import (
    ExperimentMetrics,
    TreatmentMetrics,
    HypothesisResult,
    AggregateData,
)
from .run_results import ReplicateResult
from .treatment import Treatment

__all__ = [
    "Experiment",
    "ExperimentGraph",
    "Hypothesis",
    "BaseOptimizer",
    "Objective",
    "Result",
    "ExperimentMetrics",
    "TreatmentMetrics",
    "HypothesisResult",
    "AggregateData",
    "ReplicateResult",
    "Treatment",
]



class CrystallizeError(Exception):
    """Base class for all Crystallize errors."""


class MissingMetricError(CrystallizeError):
    """Raised when a required metric is missing from the pipeline's output."""

    def __init__(self, metric: str) -> None:
        super().__init__(f"Required metric '{metric}' missing from pipeline output.")


class PipelineExecutionError(CrystallizeError):
    """Raised when a pipeline step fails unexpectedly."""

    def __init__(self, step_name: str, original_exception: Exception) -> None:
        super().__init__(f"Step '{step_name}' failed: {str(original_exception)}")
        self.original_exception = original_exception


class ContextMutationError(CrystallizeError):
    """Raised when attempting to mutate frozen context."""


class ValidationError(CrystallizeError):
    """Raised when experiment configuration is invalid."""



from .plugins import ArtifactPlugin, BasePlugin, LoggingPlugin, SeedPlugin
from .execution import AsyncExecution, ParallelExecution, SerialExecution

__all__ = [
    "ArtifactPlugin",
    "BasePlugin",
    "LoggingPlugin",
    "SeedPlugin",
    "ParallelExecution",
    "SerialExecution",
    "AsyncExecution",
]



from __future__ import annotations

from dataclasses import dataclass
import json
from pathlib import Path
from typing import Any, Callable, List, Optional, TYPE_CHECKING
import pickle
import dill

from crystallize.utils.exceptions import ContextMutationError
from crystallize.utils.constants import (
    METADATA_FILENAME,
    BASELINE_CONDITION,
)
from crystallize.plugins.plugins import ArtifactPlugin
from crystallize.datasources.datasource import DataSource

if TYPE_CHECKING:  # pragma: no cover - for type hints only
    from crystallize.utils.context import FrozenContext
    from crystallize.experiments.experiment import Experiment


@dataclass
class ArtifactRecord:
    """Container representing a file-like artifact produced by a step."""

    name: str
    data: bytes
    step_name: str


class ArtifactLog:
    """Collect artifacts produced during a pipeline step."""

    def __init__(self) -> None:
        self._items: List[ArtifactRecord] = []
        self._names: set[str] = set()

    def add(self, name: str, data: bytes) -> None:
        """Append a new artifact to the log.

        Args:
            name: Filename for the artifact.
            data: Raw bytes to be written to disk by ``ArtifactPlugin``.
        """
        if name in self._names:
            raise ContextMutationError(f"Artifact '{name}' already written in this run")
        self._names.add(name)
        self._items.append(ArtifactRecord(name=name, data=data, step_name=""))

    def clear(self) -> None:
        """Remove all logged artifacts."""
        self._items.clear()
        self._names.clear()

    def __iter__(self):
        """Iterate over collected artifacts."""
        return iter(self._items)

    def __len__(self) -> int:
        """Return the number of stored artifacts."""
        return len(self._items)


def default_loader(p: Path) -> Any:
    return p.read_bytes()


class _PickleableLoader:
    """Wrapper enabling pickling of arbitrary loader callables."""

    def __init__(self, fn: Callable[[Path], Any]) -> None:
        self.fn = fn

    def __call__(self, path: Path) -> Any:
        return self.fn(path)

    def __getstate__(self) -> bytes:
        return dill.dumps(self.fn)

    def __setstate__(self, state: bytes) -> None:
        self.fn = dill.loads(state)


def _ensure_pickleable(fn: Callable[[Path], Any]) -> Callable[[Path], Any]:
    try:
        pickle.dumps(fn)
    except Exception:
        fn = _PickleableLoader(fn)
    return fn


class Artifact(DataSource):
    """Declarative handle for reading and writing artifacts."""

    def __init__(self, name: str, loader: Callable[[Path], Any] | None = None) -> None:
        self.name = name
        self.loader = _ensure_pickleable(loader or default_loader)
        self._ctx: Optional["FrozenContext"] = None
        self._producer: Optional["Experiment"] = None
        self._manifest: Optional[dict[str, str]] = None
        self.replicates: int | None = None

    def __getstate__(self):
        """
        Customize the pickling process to exclude unpicklable attributes
        like weak references to the producer experiment.
        """
        # Copy the object's state dictionary
        state = self.__dict__.copy()

        # Remove the unpicklable weak reference before serialization
        if "_producer" in state:
            del state["_producer"]

        return state

    def _clone_with_context(self, ctx: "FrozenContext") -> "Artifact":
        clone = Artifact(self.name, loader=self.loader)
        clone._ctx = ctx
        clone._producer = self._producer
        clone._manifest = self._manifest
        clone.replicates = self.replicates
        return clone

    def write(self, data: bytes) -> None:
        if self._ctx is None:
            raise RuntimeError("Artifact not bound to context")
        self._ctx.artifacts.add(self.name, data)

    def _base_dir(self) -> Path:
        if self._producer is None:
            raise RuntimeError("Artifact not attached to an Experiment")
        plugin = self._producer.get_plugin(ArtifactPlugin)
        if plugin is None:
            raise RuntimeError("ArtifactPlugin required to load artifacts")
        # This logic is now self-sufficient and doesn't rely on plugin.version
        exp_dir = Path(plugin.root_dir) / (self._producer.name or self._producer.id)

        # Find the latest version by looking at the directories on disk
        if plugin.versioned:
            versions = [
                int(p.name[1:])
                for p in exp_dir.glob("v*")
                if p.name.startswith("v") and p.name[1:].isdigit()
            ]
            latest_version = max(versions, default=0)
        else:
            latest_version = 0

        return exp_dir / f"v{latest_version}"

    def _load_manifest(self) -> None:
        if self._manifest is not None:
            return
        base = self._base_dir()
        path = base / "_manifest.json"
        if path.exists():
            with open(path) as f:
                self._manifest = json.load(f)
        else:
            self._manifest = {}

    def fetch(self, ctx: "FrozenContext") -> Any:
        self._load_manifest()
        base = self._base_dir()
        if self.replicates is None:
            meta_path = base / METADATA_FILENAME
            if meta_path.exists():
                with open(meta_path) as f:
                    meta = json.load(f)
                self.replicates = meta.get("replicates")
        step_name = self._manifest.get(self.name)
        if step_name is None:
            raise FileNotFoundError(f"Manifest missing entry for {self.name}")

        consumer_rep = ctx.get("replicate", 0)
        rep_to_load = consumer_rep

        if self.replicates and self.replicates > 0:
            rep_to_load = consumer_rep % self.replicates

        cond = ctx.get("condition")
        path = base / f"replicate_{rep_to_load}" / cond / step_name / self.name

        if not path.exists() and cond != BASELINE_CONDITION:
            path = (
                base
                / f"replicate_{rep_to_load}"
                / BASELINE_CONDITION
                / step_name
                / self.name
            )
        if not path.exists():
            raise FileNotFoundError(f"Artifact {path} not found")
        return self.loader(path)



from __future__ import annotations

import random
from pathlib import Path
from typing import Any, Dict, Tuple

import yaml
from rich.text import Text
from textual.app import ComposeResult
from textual.containers import Container, Horizontal
from textual.screen import Screen
from textual.widgets import (
    Button,
    Footer,
    Header,
    Input,
    LoadingIndicator,
    Static,
    Tree,
)
from textual.binding import Binding

from crystallize.experiments.experiment import Experiment
from crystallize.experiments.experiment_graph import ExperimentGraph

from ..constants import ASCII_ART_ARRAY
from ..discovery import discover_configs
from ..screens.create_experiment import CreateExperimentScreen
from ..screens.run import _launch_run
from ..utils import update_replicates


class ExperimentTree(Tree):
    """Tree widget with custom binding for experiment selection."""

    BINDINGS = [b for b in Tree.BINDINGS if getattr(b, "key", "") != "enter"] + [
        Binding("enter", "run_selected", "Run", show=True)
    ]

    def action_run_selected(self) -> None:  # pragma: no cover - delegates
        screen = self.screen
        if screen is not None and hasattr(screen, "action_run_selected"):
            screen.action_run_selected()

        try:
            line = self._tree_lines[self.cursor_line]
        except IndexError:
            pass
        else:
            node = line.path[-1]
            self.post_message(Tree.NodeSelected(node))


class SelectionScreen(Screen):
    """Main screen for selecting experiments or graphs."""

    BINDINGS = [
        ("q", "quit", "Quit"),
        ("r", "refresh", "Refresh"),
        ("c", "create_experiment", "Create Experiment"),
        ("e", "show_errors", "Errors"),
    ]

    def __init__(self) -> None:
        super().__init__()
        self._load_errors: Dict[str, BaseException] = {}
        self._experiments: Dict[str, Dict[str, Any]] = {}
        self._graphs: Dict[str, Dict[str, Any]] = {}
        self._selected_obj: Dict[str, Any] | None = None

    def _update_details(self, data: Dict[str, Any]) -> None:
        """Populate the details panel with information from ``data``."""

        details = self.query_one("#details", Static)
        info = yaml.safe_load(Path(data["path"]).read_text()) or {}

        desc = info.get("description", data.get("doc", ""))
        repl = info.get("replicates", data.get("replicates", 1))
        steps = info.get("steps", [])
        treatments = list((info.get("treatments") or {}).keys())
        outputs = [
            v.get("file_name", k) if isinstance(v, dict) else str(v)
            for k, v in (info.get("outputs") or {}).items()
        ]

        details_text = [
            f"[bold]Type: {data['type']}[/bold]",
            "",
            desc,
            f"Replicates: {repl}",
        ]

        if steps:
            details_text.append("[bold]Steps:[/bold]")
            details_text.extend(f"- {s}" for s in steps)

        if treatments:
            details_text.append("[bold]Treatments:[/bold]")
            details_text.extend(f"- {t}" for t in treatments)

        if outputs:
            details_text.append("[bold]Outputs:[/bold]")
            details_text.extend(f"- {o}" for o in outputs)

        details.update("\n".join(details_text))

        rep_input = self.query_one("#replicate-input", Input)
        rep_input.value = str(repl)

    def compose(self) -> ComposeResult:
        yield Header(show_clock=True)
        yield Static(random.choice(ASCII_ART_ARRAY), id="title")
        with Container(id="main-container"):
            yield LoadingIndicator()
            yield Static(
                "Scanning for experiments and graphs...",
                id="loading-text",
            )
        yield Footer()

    async def on_mount(self) -> None:
        self.run_worker(self._discover())

    def action_refresh(self) -> None:
        self.run_worker(self._discover())

    def action_create_experiment(self) -> None:
        self.app.push_screen(CreateExperimentScreen())

    def _discover_sync(
        self,
    ) -> Tuple[
        Dict[str, Dict[str, Any]],
        Dict[str, Dict[str, Any]],
        Dict[str, BaseException],
    ]:
        """Locate ``config.yaml`` files and classify them."""

        return discover_configs(Path("."))

    async def _discover(self) -> None:
        worker = self.run_worker(self._discover_sync, thread=True)
        graphs, experiments, errors = await worker.wait()
        self._load_errors = errors
        self._experiments = experiments
        self._graphs = graphs

        main_container = self.query_one("#main-container")
        await main_container.remove_children()

        await main_container.mount(Static("Select an object to run:"))

        horizontal = Horizontal()
        await main_container.mount(horizontal)

        left_panel = Container(classes="left-panel")
        await horizontal.mount(left_panel)

        tree = ExperimentTree("root", id="object-tree")
        tree.show_root = False
        await left_panel.mount(tree)

        groups: dict[str, list[tuple[str, Dict[str, Any]]]] = {}
        for label, info in graphs.items():
            groups.setdefault(info["cli"]["group"], []).append(("Graph", info))
        for label, info in experiments.items():
            groups.setdefault(info["cli"]["group"], []).append(("Experiment", info))

        for group_name in sorted(groups):
            parent = tree.root.add(group_name, expand=True)
            items = sorted(groups[group_name], key=lambda t: t[1]["cli"]["priority"])
            for obj_type, info in items:
                label = info["label"]
                icon = info["cli"]["icon"]
                color = info["cli"].get("color")
                text = Text(f"{icon} {label} ({obj_type})")
                if color:
                    text.stylize(color)
                parent.add_leaf(
                    text,
                    {
                        "path": info["path"],
                        "label": label,
                        "type": obj_type,
                        "doc": info["description"] or "No description available.",
                    },
                )

        right_panel = Container(classes="right-panel")
        await horizontal.mount(right_panel)
        await right_panel.mount(Static(id="details", classes="details-panel"))

        btn_container = Container(id="select-button-container")
        await right_panel.mount(btn_container)
        await btn_container.mount(Button("Run", id="run-btn"))
        await btn_container.mount(Input(id="replicate-input", placeholder="replicates"))

        if self._load_errors:
            await main_container.mount(
                Static(
                    f"Failed to load {len(self._load_errors)} file(s), press e for more details",
                    id="error-msg",
                )
            )

        tree.focus()

    async def _run_interactive_and_exit(self, info: Dict[str, Any]) -> None:
        cfg = info["path"]
        obj_type = info["type"]
        try:
            if obj_type == "Graph":
                obj = ExperimentGraph.from_yaml(cfg)
            else:
                obj = Experiment.from_yaml(cfg)
        except BaseException as exc:  # noqa: BLE001
            self._load_errors[str(cfg)] = exc
            from ..screens.load_errors import LoadErrorsScreen

            self.app.push_screen(LoadErrorsScreen({str(cfg): exc}))
            return

        await _launch_run(self.app, obj)

    def action_run_selected(self) -> None:
        if self._selected_obj is not None:
            self.run_worker(self._run_interactive_and_exit(self._selected_obj))

    async def on_tree_node_highlighted(self, event: Tree.NodeHighlighted) -> None:
        if event.node.data is not None:
            data = event.node.data
            self._update_details(data)
            self._selected_obj = data
        else:
            details = self.query_one("#details", Static)
            details.update("")
            self._selected_obj = None

    async def on_tree_node_selected(self, event: Tree.NodeSelected) -> None:
        if event.node.data is not None:
            data = event.node.data
            self._update_details(data)
            self._selected_obj = data

    def action_show_errors(self) -> None:
        if self._load_errors:
            from ..screens.load_errors import LoadErrorsScreen

            self.app.push_screen(LoadErrorsScreen(self._load_errors))

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "run-btn":
            self.action_run_selected()

    def on_input_submitted(self, event: Input.Submitted) -> None:
        if event.input.id == "replicate-input" and self._selected_obj is not None:
            try:
                new_val = int(event.value)
            except ValueError:
                self.app.bell()
                return
            update_replicates(Path(self._selected_obj["path"]), new_val)
            self._selected_obj["replicates"] = new_val
            self._update_details(self._selected_obj)



import pytest
from crystallize.utils.context import FrozenContext
from crystallize.pipelines.pipeline_step import PipelineStep
from crystallize.utils.cache import compute_hash
from crystallize import pipeline_step


class AddStep(PipelineStep):
    def __init__(self, value: int) -> None:
        self.value = value

    def __call__(self, data, ctx):
        return data + self.value

    @property
    def params(self) -> dict:
        return {"value": self.value}


def test_concrete_step_basic():
    step = AddStep(2)
    ctx = FrozenContext({})
    assert step(3, ctx) == 5
    expected_hash = compute_hash({"class": "AddStep", "params": {"value": 2}})
    assert step.step_hash == expected_hash


@pipeline_step(cacheable=True)
def add(data, ctx, value: int = 1):
    return data + value


@pipeline_step(cacheable=False)
def multiply(data, ctx, factor: int):
    return data * factor


def test_pipeline_step_factory_defaults_and_hash():
    step1 = add()
    step2 = add(value=3)
    ctx = FrozenContext({})
    assert step1(1, ctx) == 2
    assert step2(1, ctx) == 4
    assert step1.cacheable is True
    assert multiply(factor=2).cacheable is False
    assert step1.step_hash == add().step_hash


def test_pipeline_step_factory_missing_param():
    with pytest.raises(TypeError):
        multiply()


@pipeline_step()
def no_ctx_step(data, *, inc: int = 1):
    return data + inc


def test_pipeline_step_without_ctx_parameter():
    ctx = FrozenContext({"inc": 2})
    step = no_ctx_step()
    assert step(3, ctx) == 5



from abc import ABC, abstractmethod
from typing import Any

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover - for type hints only
    from crystallize.utils.context import FrozenContext
    from .artifacts import Artifact


class DataSource(ABC):
    """Abstract provider of input data for an experiment."""

    @abstractmethod
    def fetch(self, ctx: "FrozenContext") -> Any:
        """Return raw data for a single pipeline run.

        Implementations may load data from disk, generate synthetic samples or
        access remote sources.  They should be deterministic with respect to the
        provided context.

        Args:
            ctx: Immutable execution context for the current run.

        Returns:
            The produced data object.
        """
        raise NotImplementedError()


class ExperimentInput(DataSource):
    """Load multiple named artifacts for an experiment."""

    def __init__(self, **inputs: "Artifact") -> None:
        if not inputs:
            raise ValueError("At least one input must be provided")
        self._inputs = inputs
        first = next(iter(inputs.values()))
        self._replicates = getattr(first, "replicates", None)
        self.required_outputs = list(inputs.values())

    def fetch(self, ctx: "FrozenContext") -> dict[str, Any]:
        return {name: art.fetch(ctx) for name, art in self._inputs.items()}

    @property
    def replicates(self) -> int | None:
        return self._replicates


# Backwards compatibility
MultiArtifactDataSource = ExperimentInput



import asyncio
from pathlib import Path

from cli.screens.run import _launch_run
from crystallize import data_source, pipeline_step
from crystallize.experiments.experiment import Experiment
from crystallize.experiments.experiment_graph import ExperimentGraph
from crystallize.pipelines.pipeline import Pipeline
from crystallize.plugins.plugins import ArtifactPlugin


@data_source
def dummy_source(ctx):
    return 0


@pipeline_step()
def add_one(data, ctx):
    return data


class FakeApp:
    def __init__(self, responses):
        self.responses = responses
        self.screens = []

    async def push_screen_wait(self, screen):
        self.screens.append(type(screen).__name__)
        return self.responses.get(type(screen).__name__)

    async def push_screen(self, screen):
        self.screens.append(type(screen).__name__)


def test_launch_run_deletes(tmp_path: Path):
    plugin = ArtifactPlugin(root_dir=str(tmp_path))
    exp = Experiment(
        datasource=dummy_source(),
        pipeline=Pipeline([add_one()]),
        name="e",
        plugins=[plugin],
    )
    exp.validate()
    graph = ExperimentGraph.from_experiments([exp])

    path = Path(plugin.root_dir) / "e"
    path.mkdir()

    app = FakeApp(
        {
            "PrepareRunScreen": ("rerun", (0,)),
            "ConfirmScreen": True,
        }
    )

    asyncio.run(_launch_run(app, graph))

    assert "RunScreen" in app.screens
    assert not path.exists()


def test_launch_run_cancel(tmp_path: Path):
    exp = Experiment(datasource=dummy_source(), pipeline=Pipeline([add_one()]))
    exp.validate()

    app = FakeApp({"PrepareRunScreen": None})
    asyncio.run(_launch_run(app, exp))

    assert "RunScreen" not in app.screens



import copy
import logging
from collections import defaultdict
from types import MappingProxyType
from typing import Any, DefaultDict, Dict, List, Mapping, Optional, Tuple

from crystallize.datasources.artifacts import ArtifactLog
from crystallize.utils.exceptions import ContextMutationError


class FrozenMetrics:
    """Immutable mapping of metric lists with safe append."""

    def __init__(self) -> None:
        self._metrics: DefaultDict[str, List[Any]] = defaultdict(list)

    def __getitem__(self, key: str) -> Tuple[Any, ...]:
        return tuple(self._metrics[key])

    def add(self, key: str, value: Any) -> None:
        self._metrics[key].append(value)

    def as_dict(self) -> Mapping[str, Tuple[Any, ...]]:
        return MappingProxyType({k: tuple(v) for k, v in self._metrics.items()})


class FrozenContext:
    """Immutable execution context shared between pipeline steps.

    Once a key is set its value cannot be modified. Attempting to do so raises
    :class:`ContextMutationError`. This immutability guarantees deterministic
    provenance during pipeline execution.

        Attributes:
            metrics: :class:`FrozenMetrics` used to accumulate lists of metric
            values.
        artifacts: :class:`ArtifactLog` collecting binary artifacts to be saved
            by :class:`~crystallize.plugins.plugins.ArtifactPlugin`.
        logger: :class:`logging.Logger` used for debug and info messages.
    """

    def __init__(
        self, initial: Mapping[str, Any], logger: Optional[logging.Logger] = None
    ) -> None:
        self._data = copy.deepcopy(dict(initial))
        self.metrics = FrozenMetrics()
        self.artifacts = ArtifactLog()
        self.logger = logger or logging.getLogger("crystallize")

    def __getitem__(self, key: str) -> Any:
        return copy.deepcopy(self._data[key])

    def __setitem__(self, key: str, value: Any) -> None:
        if key in self._data:
            raise ContextMutationError(f"Cannot mutate existing key: '{key}'")
        self._data[key] = value

    def add(self, key: str, value: Any) -> None:
        """Alias for ``__setitem__`` providing a clearer API."""
        self.__setitem__(key, value)

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        """Return the value for ``key`` if present else ``default``."""
        if key in self._data:
            return copy.deepcopy(self._data[key])
        return copy.deepcopy(default)

    def as_dict(self) -> Mapping[str, Any]:
        return MappingProxyType(copy.deepcopy(self._data))


class LoggingContext:
    """Proxy around :class:`FrozenContext` that records all key accesses."""

    def __init__(self, ctx: FrozenContext, logger: logging.Logger) -> None:
        self._ctx = ctx
        self._logger = logger
        self.reads: Dict[str, Any] = {}
        self.metrics = ctx.metrics

    # -------------------------------------------------------------- #
    def __getattr__(self, name: str) -> Any:  # pragma: no cover - passthrough
        return getattr(self._ctx, name)

    def __getitem__(self, key: str) -> Any:
        value = self._ctx[key]
        self.reads[key] = value
        self._logger.debug("Read %s -> %s", key, value)
        return value

    def __setitem__(self, key: str, value: Any) -> None:
        self._ctx[key] = value

    def add(self, key: str, value: Any) -> None:
        self._ctx.add(key, value)

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        if key in self._ctx.as_dict():
            value = self._ctx.get(key)
            self.reads[key] = value
            self._logger.debug("Read %s -> %s", key, value)
            return value
        return default

    def as_dict(self) -> Mapping[str, Any]:
        return self._ctx.as_dict()



from __future__ import annotations

from textual.app import App

from .constants import CSS
from .discovery import _import_module, _run_object, discover_objects
from .utils import _build_experiment_table, _write_experiment_summary, _write_summary
from .screens.selection import SelectionScreen

# Export these for backward compatibility
__all__ = [
    "CrystallizeApp",
    "run",
    "_import_module",
    "discover_objects",
    "_run_object",
    "_build_experiment_table",
    "_write_experiment_summary",
    "_write_summary",
]


class CrystallizeApp(App):
    """Textual application for running crystallize objects."""

    CSS = CSS

    BINDINGS = [
        ("q", "quit", "Quit"),
        ("ctrl+c", "quit", "Quit"),
    ]

    def on_mount(self) -> None:
        self.push_screen(SelectionScreen())


def run() -> None:
    CrystallizeApp().run()



import random

import numpy as np

from crystallize.datasources.datasource import DataSource
from crystallize.experiments.experiment import Experiment
from crystallize.pipelines.pipeline import Pipeline
from crystallize.pipelines.pipeline_step import PipelineStep
from crystallize.plugins.execution import ParallelExecution
from crystallize.plugins.plugins import SeedPlugin
from crystallize.utils.context import FrozenContext


def _seed_fn(seed: int) -> None:
    random.seed(seed)
    np.random.seed(seed % (2**32 - 1))


class RandomSource(DataSource):
    def fetch(self, ctx: FrozenContext) -> float:  # type: ignore[override]
        return random.random() + np.random.random()


class AddRandomStep(PipelineStep):
    def __call__(self, data, ctx):
        val = data + random.random() + np.random.random()
        ctx.metrics.add("rand", val)
        return {"rand": val}

    @property
    def params(self):
        return {}


def _run(seed_plugin, execution_plugin=None):
    pipeline = Pipeline([AddRandomStep()])
    ds = RandomSource()
    plugins = [seed_plugin]
    if execution_plugin is not None:
        plugins.append(execution_plugin)
    exp = Experiment(datasource=ds, pipeline=pipeline, plugins=plugins)
    exp.validate()
    res = exp.run(replicates=5)
    return res.metrics.baseline.metrics["rand"]


def test_seed_plugin_reproducibility_across_executors():
    seed_plugin = SeedPlugin(seed=42, seed_fn=_seed_fn)

    serial = _run(seed_plugin)
    thread = _run(seed_plugin, ParallelExecution(executor_type="thread"))
    process = _run(seed_plugin, ParallelExecution(executor_type="process"))

    assert serial == thread == process



import builtins
from crystallize.experiments.result import Result
from crystallize.experiments.result_structs import ExperimentMetrics, TreatmentMetrics


def test_result_accessors_and_errors():
    metrics = ExperimentMetrics(
        baseline=TreatmentMetrics({"a": [1]}),
        treatments={},
        hypotheses=[],
    )
    artifacts = {"model": object()}
    errors = {"run": RuntimeError("fail")}
    r = Result(metrics=metrics, artifacts=artifacts, errors=errors)
    assert r.metrics.baseline.metrics["a"] == [1]
    assert r.get_artifact("model") is artifacts["model"]
    assert r.errors == errors


def test_print_tree_plain_output(capsys):
    metrics = ExperimentMetrics(
        baseline=TreatmentMetrics({}),
        treatments={},
        hypotheses=[],
    )
    provenance = {
        "ctx_changes": {
            "baseline": {
                0: [
                    {
                        "step": "AddStep",
                        "ctx_changes": {
                            "reads": {"x": 1},
                            "wrote": {},
                            "metrics": {},
                        },
                    }
                ]
            }
        }
    }
    r = Result(metrics=metrics, provenance=provenance)
    r.print_tree()
    output = capsys.readouterr().out
    assert "AddStep" in output
    assert "x=1" in output


def test_print_tree_without_rich(monkeypatch, capsys):
    metrics = ExperimentMetrics(
        baseline=TreatmentMetrics({}),
        treatments={},
        hypotheses=[],
    )
    prov = {
        "ctx_changes": {
            "baseline": {
                0: [
                    {
                        "step": "AddStep",
                        "ctx_changes": {"reads": {"x": 1}, "wrote": {}, "metrics": {}},
                    }
                ]
            }
        }
    }
    r = Result(metrics=metrics, provenance=prov)

    original_import = builtins.__import__

    def fake_import(name, globals=None, locals=None, fromlist=(), level=0):
        if name.startswith("rich"):
            raise ImportError
        return original_import(name, globals, locals, fromlist, level)

    monkeypatch.setattr(builtins, "__import__", fake_import)
    r.print_tree()
    output = capsys.readouterr().out
    assert "Experiment Summary" in output
    assert "AddStep" in output



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Callable, List

from crystallize.plugins.plugins import BasePlugin
from crystallize.utils.constants import BASELINE_CONDITION, CONDITION_KEY, REPLICATE_KEY
from crystallize.utils.context import FrozenContext
from crystallize.pipelines.pipeline_step import PipelineStep
from crystallize.experiments.experiment import Experiment


@dataclass
class CLIStatusPlugin(BasePlugin):
    """Track progress of an experiment for the CLI."""

    callback: Callable[[str, dict[str, Any]], None]
    total_steps: int = field(init=False, default=0)
    total_replicates: int = field(init=False, default=0)
    total_conditions: int = field(init=False, default=0)
    completed: int = field(init=False, default=0)
    steps: List[str] = field(init=False, default_factory=list)

    # Add this flag
    sent_start: bool = field(init=False, default=False)

    def before_run(self, experiment: Experiment) -> None:
        # This hook is now only for internal setup, not for callbacks.
        self.completed = 0
        self.sent_start = False

    def before_replicate(self, experiment: Experiment, ctx: FrozenContext) -> None:
        # Move the 'start' event logic here, guarded by the flag
        if not self.sent_start:
            self.steps = [step.__class__.__name__ for step in experiment.pipeline.steps]
            self.total_steps = len(self.steps)
            self.total_replicates = experiment.replicates
            self.total_conditions = len(experiment.treatments) + 1
            self.callback(
                "start",
                {
                    "steps": self.steps,
                    "replicates": self.total_replicates,
                    "total": self.total_steps
                    * self.total_replicates
                    * self.total_conditions,
                },
            )
            self.sent_start = True

        # Original before_replicate logic follows
        rep = ctx.get(REPLICATE_KEY, 0) + 1
        condition = ctx.get(CONDITION_KEY, BASELINE_CONDITION)
        if condition == BASELINE_CONDITION:
            self.current_replicate = rep
        self.current_condition = condition
        self.callback(
            "replicate",
            {
                "replicate": getattr(self, "current_replicate", rep),
                "total": self.total_replicates,
                "condition": condition,
            },
        )

    def after_step(
        self,
        experiment: Experiment,
        step: PipelineStep,
        data: Any,
        ctx: FrozenContext,
    ) -> None:
        self.completed += 1
        percent = 0.0
        total = self.total_steps * self.total_replicates * self.total_conditions
        if total:
            percent = self.completed / total
        self.callback(
            "step",
            {
                "step": step.__class__.__name__,
                "percent": percent,
            },
        )



from dataclasses import dataclass
from typing import Any, Mapping, Optional, Dict, List


@dataclass
class ReplicateResult:
    """Holds the complete results from a single replicate execution."""

    baseline_metrics: Optional[Mapping[str, Any]]
    baseline_seed: Optional[int]
    treatment_metrics: Dict[str, Mapping[str, Any]]
    treatment_seeds: Dict[str, int]
    errors: Dict[str, Exception]
    provenance: Dict[str, List[Mapping[str, Any]]]



"""Screen for running experiments and graphs."""

from __future__ import annotations

import asyncio
import shutil
import sys
import traceback
from pathlib import Path
from typing import Any, Callable, List, Tuple
import queue  # Import queue

import networkx as nx
from rich.text import Text
from textual.app import App, ComposeResult
from textual.containers import VerticalScroll, Container
from textual.css.query import NoMatches
from textual.message import Message
from textual.reactive import reactive
from textual.screen import ModalScreen
from textual.widgets import Button, Header, RichLog, Static, TextArea

from crystallize.experiments.experiment import Experiment
from crystallize.experiments.experiment_graph import ExperimentGraph
from crystallize.plugins.plugins import ArtifactPlugin
from ..status_plugin import CLIStatusPlugin
from ..discovery import _run_object
from ..widgets.writer import WidgetWriter
from .delete_data import ConfirmScreen
from .prepare_run import PrepareRunScreen
from .summary import SummaryScreen


def _inject_status_plugin(
    obj: Any, callback: Callable[[str, dict[str, Any]], None]
) -> None:
    """Inject CLIStatusPlugin into experiments if not already present."""
    if isinstance(obj, ExperimentGraph):
        for node in obj._graph.nodes:
            exp: Experiment = obj._graph.nodes[node]["experiment"]
            if exp.get_plugin(CLIStatusPlugin) is None:
                exp.plugins.append(CLIStatusPlugin(callback))
    else:
        if obj.get_plugin(CLIStatusPlugin) is None:
            obj.plugins.append(CLIStatusPlugin(callback))


class RunScreen(ModalScreen[None]):
    """Display live output of a running experiment."""

    class NodeStatusChanged(Message):
        def __init__(self, node_name: str, status: str) -> None:
            self.node_name = node_name
            self.status = status
            super().__init__()

    class ExperimentComplete(Message):
        def __init__(self, result: Any) -> None:
            self.result = result
            super().__init__()

    BINDINGS = [
        ("ctrl+c", "cancel_and_exit", "Cancel and Go Back"),
        ("escape", "cancel_and_exit", "Cancel and Go Back"),
        ("q", "cancel_and_exit", "Close"),
        ("t", "toggle_plain_text", "Toggle Text"),
    ]

    node_states: dict[str, str] = reactive({})
    replicate_info: str = reactive("")
    progress_percent: float = reactive(0.0)
    step_states: dict[str, str] = reactive({})
    plain_text: bool = reactive(False)

    def __init__(self, obj: Any, strategy: str, replicates: int | None) -> None:
        super().__init__()
        self._obj = obj
        self._strategy = strategy
        self._replicates = replicates
        self._result: Any = None
        self.event_queue = queue.Queue()
        self.log_history: list[str] = []

    def watch_node_states(self) -> None:
        if not isinstance(self._obj, ExperimentGraph):
            return
        try:
            dag_widget = self.query_one("#dag-display", Static)
        except NoMatches:
            return
        text = Text(justify="center")
        order = list(nx.topological_sort(self._obj._graph))
        for i, node in enumerate(order):
            status = self.node_states.get(node, "pending")
            style = {
                "completed": "bold green",
                "running": "bold blue",
                "pending": "bold white",
            }.get(status, "bold white")
            text.append(f"[ {node} ]", style=style)
            if i < len(order) - 1:
                text.append(" ⟶  ", style="white")
        dag_widget.update(text)

    def on_node_status_changed(self, message: NodeStatusChanged) -> None:
        self.node_states = {**self.node_states, message.node_name: message.status}

    def watch_step_states(self) -> None:
        if not self.step_states:
            return
        try:
            step_widget = self.query_one("#step-display", Static)
        except NoMatches:
            return
        text = Text(justify="center")
        steps = list(self.step_states.keys())
        for i, step in enumerate(steps):
            status = self.step_states[step]
            style = {
                "completed": "bold green",
                "pending": "bold white",
            }.get(status, "bold white")
            text.append(f"[ {step} ]", style=style)
            if i < len(steps) - 1:
                text.append(" ⟶  ", style="white")
        step_widget.update(text)

    def watch_replicate_info(self) -> None:
        try:
            rep_widget = self.query_one("#replicate-display", Static)
        except NoMatches:
            return
        rep_widget.update(self.replicate_info)

    def watch_progress_percent(self) -> None:
        try:
            prog_widget = self.query_one("#progress-display", Static)
        except NoMatches:
            return
        filled = int(self.progress_percent * 20)
        bar = "[" + "#" * filled + "-" * (20 - filled) + "]"
        prog_widget.update(f"{bar} {self.progress_percent*100:.0f}%")

    def watch_plain_text(self) -> None:
        """Toggles visibility between the RichLog and the plain text TextArea."""
        try:
            log_widget = self.query_one("#live_log", RichLog)
            text_widget = self.query_one("#plain_log", TextArea)
            toggle_button = self.query_one("#toggle_text", Button)
        except NoMatches:
            return

        toggle_button.label = "Rich Text" if self.plain_text else "Plain Text"

        if self.plain_text:
            # Switched TO plain text mode
            # Join the history and load it into the TextArea
            full_log = "".join(self.log_history)
            text_widget.load_text(full_log)

            # Hide the RichLog and show the TextArea
            log_widget.display = False
            text_widget.display = True
            text_widget.focus()  # Focus the TextArea so it can be scrolled/selected
        else:
            # Switched BACK to rich text mode
            log_widget.display = True
            text_widget.display = False

    def _handle_status_event(self, event: str, info: dict[str, Any]) -> None:
        if event == "start":
            self.step_states = {name: "pending" for name in info.get("steps", [])}
            self.progress_percent = 0.0
            self.replicate_info = "Run started"
        elif event == "replicate":
            rep = info.get("replicate", 0)
            total = info.get("total", 0)
            cond = info.get("condition", "")
            self.replicate_info = f"Replicate {rep}/{total} ({cond})"
            self.step_states = {name: "pending" for name in self.step_states}
        elif event == "step":
            step = info.get("step")
            if step and step in self.step_states:
                self.step_states[step] = "completed"
            self.progress_percent = info.get("percent", 0.0)

    def compose(self) -> ComposeResult:
        with VerticalScroll(id="run-container"):
            yield Header(show_clock=True)
            yield Static(f"⚡ Running: {self._obj.name}", id="modal-title")
            yield Static("Run started", id="replicate-display")
            yield Static(id="step-display")
            yield Static(id="progress-display")
            yield Static(id="dag-display", classes="hidden")

            with Container(id="log-viewer"):
                yield RichLog(highlight=True, markup=True, id="live_log")
                yield TextArea(
                    "",
                    read_only=True,
                    show_line_numbers=False,
                    id="plain_log",
                    classes="hidden",
                )
            with Container(id="button-container"):
                yield Button("Plain Text", id="toggle_text")
                yield Button("Summary", id="summary")
                yield Button("Close", id="close_run")

    def open_summary_screen(self, result: Any) -> None:
        self.app.push_screen(SummaryScreen(result))

    def process_queue(self) -> None:
        """Polls the queue and processes events from the worker thread."""
        try:
            while not self.event_queue.empty():
                event, info = self.event_queue.get_nowait()
                self._handle_status_event(event, info)
        except queue.Empty:
            pass

    def on_mount(self) -> None:
        if isinstance(self._obj, ExperimentGraph):
            self.node_states = {node: "pending" for node in self._obj._graph.nodes}
            self.query_one("#dag-display").remove_class("hidden")
        log = self.query_one("#live_log", RichLog)

        def queue_callback(event: str, info: dict[str, Any]) -> None:
            """A simple, thread-safe callback that puts events onto a queue."""
            self.event_queue.put((event, info))

        _inject_status_plugin(self._obj, queue_callback)
        self.queue_timer = self.set_interval(1 / 15, self.process_queue)

        async def progress_callback(status: str, name: str) -> None:
            self.app.call_from_thread(
                self.on_node_status_changed, self.NodeStatusChanged(name, status)
            )

        def run_experiment_sync() -> None:
            original_stdout = sys.stdout
            sys.stdout = WidgetWriter(log, self.app, self.log_history)
            result = None
            try:

                async def run_with_callback():
                    if isinstance(self._obj, ExperimentGraph):
                        return await self._obj.arun(
                            strategy=self._strategy,
                            replicates=self._replicates,
                            progress_callback=progress_callback,
                        )
                    else:
                        return await _run_object(
                            self._obj, self._strategy, self._replicates
                        )

                result = asyncio.run(run_with_callback())

            except Exception:
                tb_str = traceback.format_exc()
                print(
                    f"[bold red]An error occurred in the worker:\n{tb_str}[/bold red]"
                )
            finally:
                sys.stdout = original_stdout
                self.app.call_from_thread(
                    self.on_experiment_complete, self.ExperimentComplete(result)
                )

        self.worker = self.run_worker(run_experiment_sync, thread=True)

    def on_experiment_complete(self, message: ExperimentComplete) -> None:
        self.process_queue()  # Process any final messages
        self._result = message.result
        try:
            if self._result is not None:
                self.open_summary_screen(self._result)
            self.query_one("#close_run").remove_class("hidden")
        except NoMatches:
            pass

    def on_unmount(self) -> None:
        """Clean up resources when the screen is removed."""
        if hasattr(self, "queue_timer"):
            self.queue_timer.stop()
        if hasattr(self, "worker") and not self.worker.is_finished:
            self.worker.cancel()

    def action_cancel_and_exit(self) -> None:
        self.app.pop_screen()

    def action_toggle_plain_text(self) -> None:
        self.plain_text = not self.plain_text

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "close_run":
            self.app.pop_screen()
        elif event.button.id == "summary":
            if self._result is not None:
                self.open_summary_screen(self._result)
        elif event.button.id == "toggle_text":
            self.action_toggle_plain_text()


async def _launch_run(app: App, obj: Any) -> None:
    selected = obj
    deletable: List[Tuple[str, Path]] = []
    if isinstance(selected, ExperimentGraph):
        for node in selected._graph.nodes:
            exp: Experiment = selected._graph.nodes[node]["experiment"]
            plugin = exp.get_plugin(ArtifactPlugin)
            if not plugin or not exp.name:
                continue
            base = Path(plugin.root_dir) / exp.name
            if base.exists():
                deletable.append((node, base))

    result = await app.push_screen_wait(PrepareRunScreen(deletable))
    if result is None:
        return
    strategy, idxs = result
    paths_to_delete = [deletable[i][1] for i in idxs]
    if paths_to_delete:
        confirm = await app.push_screen_wait(ConfirmScreen(paths_to_delete))
        if not confirm:
            return
        for p in paths_to_delete:
            try:
                shutil.rmtree(p)
            except OSError:
                pass
    await app.push_screen(RunScreen(selected, strategy, None))



"""Common string constants used across the framework."""

METADATA_FILENAME = "metadata.json"
"""Default file name for experiment metadata."""

BASELINE_CONDITION = "baseline"
"""Name used to identify the baseline condition."""

REPLICATE_KEY = "replicate"
"""Context key for the current replicate index."""

CONDITION_KEY = "condition"
"""Context key for the current treatment condition."""

SEED_USED_KEY = "seed_used"
"""Context key storing the random seed applied to a run."""

__all__ = [
    "METADATA_FILENAME",
    "BASELINE_CONDITION",
    "REPLICATE_KEY",
    "CONDITION_KEY",
    "SEED_USED_KEY",
]



from pathlib import Path

import pytest

from crystallize.datasources import Artifact
from crystallize.datasources.datasource import DataSource
from crystallize.experiments.experiment import Experiment
from crystallize.experiments.experiment_graph import ExperimentGraph
from crystallize.pipelines.pipeline import Pipeline
from crystallize.pipelines.pipeline_step import PipelineStep
from crystallize.plugins.plugins import ArtifactPlugin


class DummySource(DataSource):
    def fetch(self, ctx):
        return ctx.get("replicate", 0)


class WriteStep(PipelineStep):
    def __init__(self, artifact: Artifact, value: int = 0):
        self.artifact = artifact
        self.value = value
        self.calls = 0

    def __call__(self, data, ctx):
        self.calls += 1
        if isinstance(data, Path):
            data_val = int(data.read_text())
        else:
            data_val = data
        result = data_val + self.value
        ctx.metrics.add("val", result)
        ctx.artifacts.add(self.artifact.name, str(result).encode())
        return result

    @property
    def params(self):
        return {"value": self.value}


class ConsumeStep(PipelineStep):
    def __init__(self):
        self.calls = 0

    def __call__(self, data, ctx):
        self.calls += 1
        ctx.metrics.add("val", int(Path(data).read_text()))
        return data

    @property
    def params(self):
        return {}


def test_mixed_replicates_resume(tmp_path: Path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    plugin = ArtifactPlugin(root_dir=str(tmp_path / "arts"))

    class CountStep(PipelineStep):
        def __init__(self):
            self.calls = 0

        def __call__(self, data, ctx):
            self.calls += 1
            ctx.metrics.add("val", data)
            return data

        @property
        def params(self):
            return {}

    step_a = CountStep()
    exp_a = Experiment(
        datasource=DummySource(),
        pipeline=Pipeline([step_a]),
        plugins=[plugin],
        name="A",
        outputs=[Artifact("a.txt")],
        replicates=1,
    )

    step_b = CountStep()
    ds_b = DummySource()
    exp_b = Experiment(
        datasource=ds_b,
        pipeline=Pipeline([step_b]),
        plugins=[plugin],
        name="B",
        replicates=10,
    )

    for e in (exp_a, exp_b):
        e.validate()

    graph = ExperimentGraph()
    graph.add_experiment(exp_a)
    graph.add_experiment(exp_b)
    graph.add_dependency(exp_b, exp_a)

    graph.run()
    assert step_a.calls == 1
    assert step_b.calls == 10

    step_a2 = CountStep()
    exp_a2 = Experiment(
        datasource=DummySource(),
        pipeline=Pipeline([step_a2]),
        plugins=[plugin],
        name="A",
        outputs=[Artifact("a.txt")],
        replicates=1,
    )

    step_b2 = CountStep()
    exp_b2 = Experiment(
        datasource=ds_b,
        pipeline=Pipeline([step_b2]),
        plugins=[plugin],
        name="B",
        replicates=10,
    )

    for e in (exp_a2, exp_b2):
        e.validate()

    graph2 = ExperimentGraph()
    graph2.add_experiment(exp_a2)
    graph2.add_experiment(exp_b2)
    graph2.add_dependency(exp_b2, exp_a2)
    res = graph2.run(strategy="resume")

    assert step_a2.calls == 0
    assert step_b2.calls == 0
    assert res["B"].metrics.baseline.metrics["val"] == list(range(10))


class PassStep(PipelineStep):
    def __call__(self, data, ctx):
        return data

    @property
    def params(self):
        return {}


@pytest.mark.asyncio
async def test_progress_callback_resume():
    """Test that progress callbacks are invoked during graph execution."""
    exp_a = Experiment(
        datasource=DummySource(), pipeline=Pipeline([PassStep()]), name="a"
    )
    exp_a.validate()

    graph = ExperimentGraph()
    graph.add_experiment(exp_a)

    callback_events = []

    async def progress_cb(status: str, name: str):
        callback_events.append((status, name))

    await graph.arun(progress_callback=progress_cb)

    assert callback_events == [("running", "a"), ("completed", "a")]
